from pathlib import Path
from typing import Optional, Tuple
from scipy.stats import circmean

import numpy as np
import pandas as pd


####    Private UTIL FUNCTIONS    ####
def __csv_file_exists(file_path: Path) -> bool:
    """
    Checks if a Path is a valid path to an existing .CSV file.
    
    Args:
        file_path: The Path to the file being checked.
    
    Returns:
        True if the csv file at the given Path already exists and False otherwise.

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file is not a csv file.
        FileNotFoundError: If the parent directory of the file does not exist.

    """
    # Check Parameters for validity
    if not isinstance(file_path, Path):
        raise TypeError("file_path must be a pathlib.Path object.")
    if file_path.suffix.lower() != ".csv":
        raise ValueError("The filename must end with '.csv'.")
    if not file_path.parent.exists():
        raise FileNotFoundError(f"The directory '{file_path.parent}' does not exist.")
    
    # Return if the file path exists
    return file_path.exists()

# Data File Functions
def load_data_csv(
        file_path: Path, 
        date_time_column: str,
        date_time_format: str = "%m/%d/%Y %I:%M %p",
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> pd.DataFrame:
    '''
    Loads data from a csv file into a DataFrame while filtering and cleaning the data.

    Loads data from a csv file into a DataFrame. Strips the column names of whitespace on either end and removes apostrophes.
    Filters data by the given dates. If no start or end date is specified the oldest and newest dates in the data are used respectively.
    Converts the date time column to Date Time objects and numeric columns to Numeric Data Types
    Adds columns for the date_norm, DOY, and Year.
        Column names can be found by running AEM.settings.calc_calculation_labels().
        date_norm stores the date normalized to contain the same time, midnight.
        doy stores the number of days since January 1st, where January 1st = 0 and December 31st = 364 or 365 during a leap year.
        year stores the current year.

    Args:
        file_path: The path to your csv data file.
        start_date: Optional. The date to start filtering from (Inclusive).
        end_date: Optional. The date to stop filtering on (Inclusive).
        date_format: The date_format to use on the file.
    
    Returns:
        A DataFrame with the information from the csv file filtered by the specified dates.

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.csv'.
        FileNotFoundError: If the file could not be found.
    '''

    # Check Parameters
    if not __csv_file_exists(file_path):
        raise FileNotFoundError(f"The file, {file_path},  that you are trying to load does not exist.")
    if start_date and end_date and pd.to_datetime(start_date) > pd.to_datetime(end_date):
        raise ValueError("The end date must be after the start date.\nStart Date:\t{start_date}\nEnd Date:\t{end_date}")
    
    # Read the data from the given csv file and make the formatting consistent
    df = None
    df = pd.read_csv(file_path)
    df.columns = df.columns.str.strip().str.replace("'", "")

    if date_time_column not in df.columns:
        raise ValueError(f"The given date column name, {date_time_column}, was not found in the csv data!")
    

    df[date_time_column] = pd.to_datetime(df[date_time_column], format=date_time_format)



    # Filter data using the start and end dates
    if start_date:
        df = df[df[date_time_column] >= pd.to_datetime(start_date)]
    if end_date:
        df = df[df[date_time_column] <= pd.to_datetime(end_date)]

    # Adds a Year, DOY, and Normalized Date column to the df
    df['DOY'] = df[date_time_column].dt.dayofyear-1
    df['YEAR'] = df[date_time_column].dt.year
    df['NORMALIZED_DATE'] = pd.to_datetime(df[date_time_column], format=date_time_format).dt.normalize()


    return df.reset_index(drop=True)

def interpolate_df(
        df: pd.DataFrame, 
        col_names: Optional[list[str]] = None,
        method: str = "linear"
        ) -> pd.DataFrame:
    """
    Interpolates missing data within a DataFrame.

    Interpolates data using dataframes interpolate function. 
    If a list of column names are provided only those columns are interpolated.

    Args:
        df: The DataFrame to interpolate
        col_names: An optional list of columns names to interpolate the data for
        method: The dataframe interpolation method to use

    Returns:
        The DataFrame.

    Raises:
        KeyError: If one of the col_names was not found in the dataframe
    """
    if col_names:
        if len(col_names) == 0:
            raise ValueError("label_keys must containa list of keys or be none")
        for col in col_names:
            if col not in df.columns:
                raise KeyError(col + " was not found in the df")
            df[col] = df[col].interpolate(method=method)
    else:
        for col in df.columns.values:
            df[col] = df[col].interpolate(method=method)
    
    return df

def save_data_csv(
        df: pd.DataFrame,
        file_path: Path
    ) -> Path:
    """
    Saves your DataFrame to the given csv file.

    Args:
        df: The dataframe to be saved.
        file_path: A Path to where you would like to save the data.
    
    Returns:
        The Path to the newly saved file.

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.csv'.
        FileNotFoundError: If the parent directory does not exist.
        FileExistsError: If the file already exists.
    """

    # Check Parameters
    if __csv_file_exists(file_path):
        print(f"The file {file_path} already exists.")
        usr_input = input("Are you sure you want OVERWRITE the file(y/n):  ")
        if not usr_input or usr_input != "y":
            raise FileExistsError(f"The file, {file_path}, already exists.")
    
    # Save DataFrame
    df.to_csv(file_path, index=False)

    return file_path


# Fletcher's Functions
def match_datetimes(
        target_dt: np.ndarray,
        desired_dt: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Matches each datetime in 'desired_dt' to the closest datetime in 'target_dt'.

    Args:
        target_dt: The np array of actual weather date_times
        desired_dt:    The np array of date_times to find the closest match for
    
    Returns:
        A tuple of 4 numpy arrays
        The first contains the original desired_dt
        The second contains the date_times matched from the target_dt.
        The third contains the indices of the matched date_times in the original target_dt
        The fourth contains the the difference in time between the desired and matched times
    """
    idx = np.searchsorted(target_dt, desired_dt, side="left")
    idx = np.minimum(idx, len(target_dt) - 1)

    for i in range(len(idx)):
        if idx[i] > 0:
            before = target_dt[idx[i] - 1]
            after = target_dt[idx[i]]
            if abs(desired_dt[i] - before) < abs(desired_dt[i] - after):
                idx[i] -= 1

    matched_times = target_dt[idx]
    diffs = np.abs(desired_dt - matched_times)

    return desired_dt, matched_times, idx, diffs
 
def df_as_dict(
        df: pd.DataFrame,
        cols: Optional[np.ndarray],
        idx: Optional[np.ndarray]
    ) -> dict[str, list[float]]:
    """
    Extracts values from a DataFrame into a dictionary

    Allows users to create a dictionary from a Dataframe. Column names are used for the dictionary keys.
    Only the columns the user provides are added to the dictionary. If no columns are provided all columns are added.
    The value stored in each key is a list of data from the given DataFrame column
    Only the data for 

    Args:
        df:   The DataFrame containing weather data
        cols: The key names to be used in the dictionary
        idx:  The indices in the DataFrame to be used in the dictionary

    Returns:
        A dictionary of {column_name: values}.
    """
    # Get all Columns if None are specified
    if cols is None or len(cols) == 0:
        cols = np.asarray(df.columns)
    # Get all indices if None are specified
    if idx is None or len(idx) == 0:
        idx = np.asarray([i for i in range(df.shape[0])])

    return_dict = {}
    for col in cols:
        if col in df.columns:
            return_dict[col] = df[col].iloc[idx].tolist()
    return return_dict






from pathlib import Path
from typing import Optional
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt


from agroecometrics import equations

# Gets the acutal labels of columns based on the user settings


####    UTIL FUNCTIONS    ####

def __check_png_filename(file_path: Path):
    """
    Validates that the provided file path ends with '.png' and that the directory exists.

    Args:
        file_path: A Path object representing the output file path.

    Returns:
        True if the file_path is a png file whose parent folder exists
    
    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    if not isinstance(file_path, Path):
        raise TypeError("file_path must be a pathlib.Path object.")
    if file_path.suffix.lower() != ".png":
        raise ValueError("The filename must end with '.png'.")
    if file_path.parent and not file_path.parent.exists():
        raise FileNotFoundError(f"The directory '{file_path.parent}' does not exist.")
    return True

def __save_plot(file_path: Path) -> Path:
    """
    Prepares plot to be saved and saves it.

    Adds labels to plt figure if they exist, saves the plot to the given file_path, and
        closes the plt plot after saving

    Args:
        file_path: A Path object representing the output file path.

    Returns:
        The resolved file path where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Validate input parameters
    __check_png_filename(file_path)
    
    handles, labels = plt.gca().get_legend_handles_labels()
    if handles:  # Only add legend if there are labeled elements
        plt.legend()
    plt.savefig(file_path, dpi=300, bbox_inches="tight", pad_inches=0.4)
    plt.close()

    return file_path.resolve()


####    AIR TEMPERATURE PLOTS    ####

def plot_air_temp(
        air_temps: np.ndarray,
        predicted_temps: np.ndarray,
        date_times: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a plot of air temperature over the time.

    Args:
        air_temps: The actually air temperatures.
        predicted_temps: The predicted air temperatures.
        date_times: The datetime objects corresponding to the each actual and predicted temp.
        file_path: A Path object representing the output file path.

    Returns: 
        The resolved file path where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    global LABELS
    __check_png_filename(file_path)

    # Plot actual vs predicted temperature
    plt.figure(figsize=(8, 4))
    plt.scatter(date_times, air_temps, s=5, color="gray", label="Observed")
    plt.plot(date_times, predicted_temps, label="Predicted", color="tomato", linewidth=1)
    plt.ylabel("Air temperature (Celsius)")
    plt.xlabel("Date")
    
    return __save_plot(file_path)


####    SOIL TEMPERATURE PLOTS    ####

def plot_yearly_soil_temp(
        soil_temp: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a plot of modeled soil temperature over a year's time

    Args:
        T_soil: A numpy array of soil temperatures for each day of the year in order
        file_path: A Path object representing the output file path.

    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Check Parmeters
    __check_png_filename(file_path)

    # Create the soil temperature plot
    doy = np.arange(0,365)
    plt.figure()
    plt.plot(doy,soil_temp)
    plt.ylabel("Surface Soil Temperature (Celsius)")
    plt.xlabel("Day of Year")
   
    return __save_plot(file_path)

def plot_daily_soil_temp(
        soil_temps: np.ndarray,
        soil_depths: int,
        file_path: Path,
    ) -> Path:
    """
    Creates a plot of modeled soil temperature at a given depth

    Args:
        soil_temp: Is the predicted soil temperatures (°C)
        depth: Is the depth that the soil temperature predictions are made at. (m)
        file_path: A Path object representing the output file path.

    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Check Parameters
    __check_png_filename(file_path)

    # Create Plot
    plt.figure()
    plt.plot(soil_depths, soil_temps)
    plt.ylabel("Soil temperature (Celsius)")
    plt.xlabel("Soil Depth (Centimeters)")
    
    return __save_plot(file_path)

def plot_yearly_3d_soil_temp(
        doy_grid: np.ndarray,
        depth_grid: np.ndarray, 
        temp_grid: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a 3d plot of soil temperature at different depths over the course of a year

    Args:
        doy_gird: A 2d np.ndarray with shape (Nz, 365) containing the day of year
                for each plot point.
        z_grid:   A 2d np.ndarray with shape (Nz, 365) containing the soil depth
                for each plot point.
        t_grid:   A 2d np.ndarray with shape (Nz, 365) containing the soil temperature
                for each plot point.
        file_path: A Path object representing the output file path.
    
    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Check Parameters
    __check_png_filename(file_path)

    # Create Plot
    fig = plt.figure(figsize=(10, 6), dpi=80, constrained_layout=True) # 10 inch by 6 inch dpi = dots per inch
    ax = fig.add_subplot(111, projection="3d")
    surf = ax.plot_surface(doy_grid, depth_grid, temp_grid, cmap="viridis", antialiased=False)
    fig.colorbar(surf, shrink=0.5, aspect=20)
    frame = surf = ax.plot_wireframe(doy_grid, depth_grid, temp_grid, linewidth=0.5, color="k", alpha=0.5)

    # Label x,y, and z axis
    ax.set_xlabel("Day of the year")
    ax.set_ylabel("Soil depth [cm]")
    ax.set_zlabel("Soil temperature (Celsius)")

    # Set position of the 3D plot
    ax.view_init(elev=30, azim=35)

    return __save_plot(file_path)

def plot_soil_temp_predictions(
        air_temp: np.ndarray,
        pred_temp: np.ndarray,
        depth: float,
        file_path: Path,
        soil_temp: Optional[np.ndarray] = None,
    ) -> Path:
    """
    Creates a plot of air temperature and the predicted soil temperature on a particular date

    Creates a plot showing the given air temperature and the predicted soil temperature
        for the given depth. Optionally plots the actual soil temperature for comparision

    Args:
        air_temp: The air temperature collected every 5 minutes
        pred_temp: The soil temperature prediction given in 5 minute intervals
        depth: The depth that the soil temperature was predicted at
        file_path: A Path object representing the output file path.
        soil_temp: The collected soil temperature given in 5 minute intervals

    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or if the needed label(s) isn't found in the df
        FileNotFoundError: If the parent directory does not exist.
    """    
    # Validate input parameters
    __check_png_filename(file_path)
    
    # Generate time intervals
    time_passed = np.arange(0, 1440, 5)
    time = time_passed / 60 

    # Create Plot
    plt.figure(figsize=(8, 4))
    plt.scatter(time, air_temp, s=5, color="gray", label="Air Temperature")
    plt.plot(time, pred_temp, label=f"Predicted: {depth}in", color="tomato", linewidth=1)

    # Add Optional actual soil temperature
    if soil_temp is not None:
        plt.scatter(time, soil_temp, s=5, color="blue", label="Soil Temp")

    # Set labels
    plt.ylabel("Temperature  (Celsius)")
    plt.xlabel("Time (Hours)")
    plt.xticks(ticks=[0, 6, 12, 18, 24], labels=["12 AM", "6 AM", "12 PM", "6 PM", "12 AM"])

    return __save_plot(file_path)

def plot_3d_soil_temp_predictions(
        time_grid: np.ndarray,
        depth_grid: np.ndarray,
        temp_grid: np.ndarray,
        file_path: Path,
    ) -> Path:
    """
    Creates a 3D plot of predicted soil temperature over the course of a single day at different depths.

    Args:
        time_grid: A 2D numpy array of shape (n_depths, n_times), where each value is time in minutes.
        depth_grid: A 2D numpy array matching time_grid, where each value is the depth in cm.
        temp_grid: A 2D numpy array of predicted soil temperatures (same shape as time_grid).
        file_path: A Path object representing the output file path.

    Returns:
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or if the needed label(s) isn't found in the df
        FileNotFoundError: If the parent directory does not exist.
    """
    __check_png_filename(file_path)

    # Convert minutes to hours for plotting
    hours_grid = time_grid / 60
    depth_grid = -depth_grid  # Flip so depth increases downward

    # Create figure and axis
    fig = plt.figure(figsize=(10, 6), dpi=100, constrained_layout=True)
    ax = fig.add_subplot(111, projection='3d')

    # Surface plot
    surf = ax.plot_surface(hours_grid, depth_grid, temp_grid, cmap="viridis", antialiased=False)
    fig.colorbar(surf, shrink=0.5, aspect=20)

    # Optional: wireframe overlay
    ax.plot_wireframe(hours_grid, depth_grid, temp_grid, linewidth=0.5, color='k', alpha=0.3)

    # Axis labels
    ax.set_xlabel("Hour of Day")
    ax.set_ylabel("Soil Depth [cm]")
    ax.set_zlabel("Soil Temperature (°C)")

    # Set viewing angle
    ax.view_init(elev=30, azim=35)

    # Save and return path
    return __save_plot(file_path)


####    EVAPOTRANSPIRATION PLOTS    ####

def plot_evapo_data(
        evapotranspiration: np.ndarray,
        date_times: np.ndarray,
        file_path: Path,
        model_label: str="EvapoTranspiration",
    ) -> Path:
    """
    Creates a plot of different evapotraspiration data over time

    Creates a plot over time of the predicted evapotranspiration in mm/day.
      Creates labels and uses different colors for each of the models. 
      Creates a legend of the model names and colors.

    Args:
        evapotranspiration: The evapotranspiration (mm/day)
        date_times: The datetime objects corresponding to the each actual and predicted temp.
        file_path: A Path object representing the output file path.
        model_label: Is the label to use on the x-axis for the model
    
    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or if the needed label(s) isn't found in the df
        FileNotFoundError: If the parent directory does not exist.
    """
    __check_png_filename(file_path)


    # Generates a new Plot
    plt.figure(figsize=(10,4))

    plt.plot(date_times, evapotranspiration, label=model_label)
    
    # Adds plot label
    plt.ylabel("Evapotranspiration (mm/day)")
    plt.xlabel("Date")
    
    return __save_plot(file_path)


####    RAILFALL PLOTS    ####

def plot_rainfall(
        rainfall: np.ndarray,
        runoff: np.ndarray,
        date_times: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a plot of rainfall and runoff over time.

    Args:
        df: DataFrame with cumulative rainfall and runoff
        file_path: A Path object representing the output file path.

    Returns: 
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or the rainfall, runnoff, and date_times parameters are not the same length
        FileNotFoundError: If the parent directory does not exist.
    """
    __check_png_filename(file_path)

    # Create plot with rain and runoff data
    plt.figure(figsize=(6, 4))
    plt.plot(date_times, rainfall, color="blue", label="Rainfall")
    plt.plot(date_times, runoff, color="red", label="Runoff")
    plt.ylabel("Rainfall or Runoff (mm)")

    return __save_plot(file_path)


####    GROWING DEGREE DAYS PLOTS    ####

def plot_gdd(
        gdd: np.ndarray,
        date_times: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a plot of the Growing Degree Days that occured over each time segment in the data

    Args:
        gdd: The Growing Degree Days that occured 
        date_times: The datetime objects corresponding to the each actual and predicted temp.
        file_path: A Path object representing the output file path.

    Returns:
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or if the needed label(s) isn't found in the df
        FileNotFoundError: If the parent directory does not exist.
    """
    __check_png_filename(file_path)

    # Create plot
    plt.figure(figsize=(6, 4))
    plt.plot(date_times, gdd)
    plt.xlabel("Date")
    plt.ylabel(f'Growing degree days {chr(176)}C-d)')

    return __save_plot(file_path)
    
def plot_gdd_sum(
        gdd_sum: np.ndarray,
        date_times: np.ndarray,
        file_path: Path
    ) -> Path:
    """
    Creates a plot of the Cumulative Growing Degree Days that occured over the data

    Args:
        gdd_sum: The cummulative Growing Degree Days that have occured since the start of the data.
        date_times: The datetime objects corresponding to the each actual and predicted temp.
        file_path: A Path object representing the output file path.

    Returns:
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png' or if the needed label isn't found in the df
        FileNotFoundError: If the parent directory does not exist.
    """
    __check_png_filename(file_path)

    # Create plot
    plt.figure(figsize=(6,2))
    plt.plot(date_times, gdd_sum)
    plt.xlabel("Date")
    plt.ylabel(f'Growing degree days sum {chr(176)}C-d)')

    return __save_plot(file_path)
    

####    PHOTOPERIOD PLOTS    ####

def plot_yearly_photoperiod(lat: float, file_path: Path):
    """
    Creates a plot of the photoperiod at a specified latitude over a year's time. Not accurate near polar regions.

    Args:
        latitude: Latitude in decimal degress. Where the northern hemisphere is .
            positive and the southern hemisphere is negative.
        file_path: A Path object representing the output file path.

    Returns:
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Check Parameters
    __check_png_filename(file_path)

    # Set up plot with Title and axes
    doy = np.arange(0,366)
    plt.figure(figsize=(6,4))
    plt.title('Latitude:' + str(lat))
    plt.xlabel('Day of the year', size=14)
    plt.ylabel('Photoperiod (hours per day)', size=14)

    # Calulate photoperiods and adds them to the plot
    photoperiods, __, __, __, __, __ = equations.photoperiod_at_lat(lat, doy)

    plt.plot(doy, photoperiods, color='k')

    return __save_plot(file_path)

def plot_daily_photoperiod(doys: np.ndarray, file_path: Path):
    """
    Creates a plot of the photoperiod from -45 to 45 degree latitude on the given day of year

    Args:
        doys: A np.ndarray of the days of year (0-365) where January 1st is 0 and 365 to perform the calculation
        file_path: A Path object representing the output file path.

    Returns:
        The filename where the plot was saved

    Raises:
        TypeError: If file_path is not a Path object.
        ValueError: If the file extension is not '.png'.
        FileNotFoundError: If the parent directory does not exist.
    """
    # Check Parameters
    __check_png_filename(file_path)

    # Set up plot with Title and axes
    plt.figure(figsize=(6,4))
    plt.xlabel('Latitude (decimal degrees)', size=14)
    plt.ylabel('Photoperiod (hours per day)', size=14)

    # Calculated photoperiods and adds them to the plot
    latitudes = np.linspace(-45, 45, num=180)

    # Loop over each day and plot
    for doy in doys:
        photoperiod, *_ = equations.photoperiod_on_day(latitudes, doy)
        plt.plot(latitudes, photoperiod, label=f'DOY {doy}')
    
    return __save_plot(file_path)










from typing import Optional, Tuple
from scipy.stats import circmean

import numpy as np
import pandas as pd

import agroecometrics as AEM


# Constants Representing the number of seconds in a Day and Year
DAY_SECONDS = 24*60*60
YEAR_SECONDS = DAY_SECONDS*365

# Helper Functions
def __compute_vapor_sat_pressure(temp: np.ndarray) -> np.ndarray:
    """
    Computes saturation vapor pressure based Tetens formula.
    
    Args:
        temp: Temperature used to calculate the saturation vapor pressure (°C).
    
    Returns:
        Computed saturation vapor pressure (kPa).
    """
    e_sat = 0.6108 * np.exp(17.27 * temp/(temp+237.3)) 
    return e_sat

def __compute_solar_radiation(doy: np.ndarray, lat: float) -> np.ndarray:
    """
    Computes extra-terrestrial solar radiation using the FAO Penman-Monteith method.
    
    Args:
        doy: Day of year (0-365) where January 1st is 0 and 365.
        lat: Latitude in decimal degress. Where the northern hemisphere is positive and the southern hemisphere is negative.

    Returns:
        Extra-terrestrial solar radiation (Ra) in MJ/m²/day.
    """
    lat_radians = np.pi / 180 * lat 
    dr = 1 + 0.033 * np.cos(2 * np.pi * doy/365) # Inverse relative distance Earth-Sun
    solar_declination = 0.409 * np.sin((2 * np.pi * doy/365) - 1.39) # Solar delcination
    Gsc = 0.0820 # Solar constant

    sunset_hour_angle = np.arccos(-np.tan(lat_radians) * np.tan(solar_declination)) # Sunset hour angle

    Ra = 24 * 60 / np.pi * Gsc * dr
    Ra = Ra * (sunset_hour_angle * np.sin(lat_radians) * np.sin(solar_declination) + np.cos(lat_radians) * np.cos(solar_declination) * np.sin(sunset_hour_angle))

    return Ra

# Air Temperature model
def model_air_temp(air_temps: np.ndarray, date_times: np.ndarray) -> np.ndarray:
    '''
    Creates an air temperature model and creates air temperature estimates using model.

    Creates an air temperature model by finding the best fit for a sinusoidal function.
        Estimates Parameters using the given data.
        Creates a temperature estimate for each day in the df using the model

    Args:
        air_temps: The daily average air temperature (°C)
        date_times: The datetime corresponding to each air temperature measurement
    
    Returns: 
        A numpy array of predicted daily temperatures over the period of the dataframe
    '''
    # Estimate sinusoidal parameters
    avg_temp = np.mean(air_temps)    # Mean Temperature | Units: °C
    thermal_amp = (np.max(air_temps) - np.min(air_temps)) / 2  # | Units: °C

    # Use the day with the minimum and maximum temperatures to calulate time_laf
    min_idx = np.argmin(air_temps[:int(len(air_temps) / 2)])
    min_date = pd.to_datetime(date_times[min_idx])
    max_idx = np.argmax(air_temps)
    max_date = pd.to_datetime(date_times[max_idx])
    time_lag_date = min_date + (max_date - pd.Timedelta(days=183) - min_date) / 2
    time_lag = (time_lag_date - pd.Timestamp(time_lag_date.year, 1, 1)).days  # days since Jan 1

    # Convert all dates to day of year
    date_times = pd.to_datetime(date_times)
    doys = date_times.dt.dayofyear

    # Generate sinusoidal temperature predictions
    pred_temp = avg_temp + thermal_amp * np.cos(
        2 * np.pi * ((doys - time_lag) / 365.0) + np.pi
    )

    return np.asarray(pred_temp)


# Soil Temperature models
def yearly_soil_temp(
        depth: int,
        avg_temp: int=25,
        thermal_amp: int = 10,
        thermal_diffusivity: float = 0.203,
        time_lag: int = 15
    ) -> np.ndarray:
    """
    Models soil temperature over one year at the given depth.

    Creates a model to estimate soil temperature given a day of year and depth.
        Predicts the soil temperature everyday at the specified depth.

    Args:
        depth: The depth to model the soil temperature (m).
        avg_temp: The annual average surface temperature (°C).
        thermal_amp: The annual thermal amplitude of the soil surface (°C).
        thermal_dif: The Thermal diffusivity obtained from KD2 Pro instrument (mm^2 / s).
        time_lag: The time lag between Jaunary 1st and the coldest day of the year (days).

    Return:
        Predicted soil temperature at the given depth for each day of the year (°C).
    """
    # Set Constants
    phase_frequency = 2 * np.pi / YEAR_SECONDS # Phase Frequency | Units: 1 / s
    PHASE = np.pi/2 + phase_frequency * (time_lag * DAY_SECONDS) # Phase constant | Units: None

    # Calculate Damping Depth
    thermal_diffusivity = thermal_diffusivity / 100    # Unit Conversion | Units: cm^2 / s
    damp_depth = (2*thermal_diffusivity/phase_frequency)**(1/2) # Units: cm

    # Estimate the temperatures for each day of the year
    doy = np.arange(0,365)*DAY_SECONDS # Units: s
    soil_temp = avg_temp + \
        thermal_amp * np.exp(-depth / damp_depth) * \
        np.sin(phase_frequency * doy - depth / damp_depth - PHASE)
    
    return  np.asarray(soil_temp)

def daily_soil_temp(
        doy: int,
        max_depth: int,
        interpolations: int = 100,
        avg_temp: int = 25,
        thermal_amp: int = 10,
        thermal_diffusivity: int = 0.203,
        timelag: int = 15
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Models soil temperature on a particular day of the year

    Creates a model to estimate soil temperature given a day of year and depth.
        Calculates depths to use by linearly interpolating the max_depth by the number of interpolations.
        Uses the model to estimate the soil temperature at a range of depths on the specified day of year.
    
    Args:
        doy: Day of year (0-365) where January 1st is 0 and 365.
        max_depth: The maximum depth in centimeters to model the soil temperature.
        interpolations: The number of interpolated depths to caluculate soil temperature at.
        avg_temp: The annual average surface temperature (°C).
        thermal_amp: The annual thermal amplitude of the soil surface (°C).
        thermal_diffusivity: The Thermal diffusivity obtained from KD2 Pro instrument [mm^2/s].
        time_lag: The time lag in days (0-365) where January 1st is 0 and 365.

    Returns:
        A tuple containing two numpy arrays.
            The first array contains the soil temperature predictions (°C)
            The second array contians the depth of each prediction (cm)
    """
    # Set Constants
    phase_frequency = 2 * np.pi / (YEAR_SECONDS) # Phase Frequency | Units: 1 / s
    PHASE = np.pi / 2 + phase_frequency * (timelag * DAY_SECONDS) # Phase constant | Units: None

    # Calculate Damping Depth
    thermal_diffusivity = thermal_diffusivity / 100    # Unit Conversion | Units: cm^2 / s
    damp_depth = (2 * thermal_diffusivity / phase_frequency) ** 0.5 # Units: cm

    # Estimate the temperatures for each depth
    soil_depths = np.linspace(max_depth/interpolations, max_depth, interpolations) # Interpolation depths | Units: cm
    soil_temp = avg_temp + \
        thermal_amp * np.exp(-soil_depths / damp_depth) * \
        np.sin(phase_frequency * doy - soil_depths / damp_depth - PHASE)

    return (np.asarray(soil_temp), np.asarray(soil_depths))

def yearly_3d_soil_temp(
        max_depth: int,
        interpolations: int = 1000,
        avg_temp: int = 25,
        thermal_amp: int = 10,
        thermal_diffusivity: int = 0.203,
        timelag: int = 15
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Models soil temperature over a full year (0–365) and across depth.

    Creates a model to estimate soil temperature given a day of year and depth.
        Calculates depths to use by linearly interpolating the max_depth by the number of interpolations
        Uses the model to create a matrix of estimations for each day of the year at each depth.
        
    Args:
        max_depth: The maximum depth in centimeters to model the soil temperature
        interpolations: The number of interpolated depths to calculate soil temperature at
        avg_temp: The annual average surface temperature (°C)
        thermal_amp: The annual thermal amplitude of the soil surface (°C)
        thermal_diffusivity: The Thermal diffusivity obtained from KD2 Pro instrument [mm^2/s]
        timelag: The time lag in days (0-365) where January 1st is 0 and 365
    
    Returns:
        A tuple of (doy_grid, z_grid, temp_grid), where each is a 2D NumPy array. 
        Each grid represents a Matrix of depths and doys.
        doy_grid varies the day of year between columns. Each columns represents the same doy.
        depth_grid varies the depth between rows. Each row represents the same depth
        temp_grid varies between rows and columns. 
        Each point represents the predicted temperature for the doy and depth indicated by the same point in the doy_grid and depth_grid respectively
    """
    # Set Constants
    phase_frequency = 2 * np.pi / 365 # Phase Frequency | Units: 1 / s
    PHASE = np.pi/2 + phase_frequency*timelag # Phase constant | Units: None

    thermal_diffusivity = thermal_diffusivity * 0.01 * 86400 # Unit conversion | Units: cm^2/day
    damp_depth = (2*thermal_diffusivity/phase_frequency)**(1/2) # Damping depth 

    doys = np.arange(1,366)
    depths = np.linspace(0, max_depth, interpolations) # Interpolation depths
    doy_grid, depth_grid = np.meshgrid(doys,depths)
    
    temp_grid = avg_temp + thermal_amp * np.exp(-depth_grid/damp_depth) * np.sin(phase_frequency*doy_grid - depth_grid/damp_depth - PHASE)

    return doy_grid, depth_grid, temp_grid

def model_soil_temp(
        air_temps: np.ndarray,
        depth: float,
        thermal_diffusivity: int = 0.203,
    ) -> np.ndarray:
    """
    Creates soil temperature predictions for each air temperature provided over the course of a day

    Estimates the parameters for a sinusodial function modeling the daily surface temperature.
        Uses the parameter estimations to create a model of soil temperatures at different depths.
        Predicts the soil temperatue at the given depth for each air temperature provided

    Args:
        air_temps: The soil surface air temperatures for a given day (°C).
        depth: The depth to model the soil temperature (cm).
        thermal_diffusivity: The Thermal diffusivity obtained from KD2 Pro instrument [mm^2/s].

    Returns:
        A numpy array containing the predicted temperatures (°C) every 5 minutes 
            starting at midnight at the specified depth and a numpy array containing
            the given air temperatures
    """
    # Constants
    phase_frequency = 2 * np.pi / DAY_SECONDS # Phase Frequency | Units: 1 / s

    # Calculate Damping Depth
    thermal_diffusivity = thermal_diffusivity / 100 # Unit Conversion: cm^2 / s
    damp_depth = (2 * thermal_diffusivity / phase_frequency) ** 0.5 # Units: cm

    measurement_offset = DAY_SECONDS/len(air_temps)

    # Estimate sinusoidal parameters
    avg_temp = np.mean(air_temps)    # Units: °C
    thermal_amp = (max(air_temps) - min(air_temps)) / 2 # Units: °C

    # Find TimeLag
    min_idx = np.argmin(air_temps)
    timelag = min_idx*measurement_offset

    # Generate predictions for every air temperature provided
    times = [i*measurement_offset for i in len(air_temps)] # Units: s
    temp_predictions = avg_temp + \
        thermal_amp * np.exp(-depth / damp_depth) * \
        np.sin(phase_frequency * (times - timelag) - (depth / damp_depth))

    return temp_predictions

def model_3d_soil_temp(
        air_temps: np.ndarray,
        max_depth: float,
        interpolations: int = 1000,
        thermal_diffusivity: float = 0.203
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Models soil temperature over a full day (in 5-minute intervals) across depth 
    using air temperature data to parameterize a sinusoidal model.

    Args:
        air_temps: The soil surface air temperatures for a given day (°C).
        max_depth: Maximum depth to model the soil temperature (cm).
        interpolations: Number of interpolated depth points.
        thermal_diffusivity: Thermal diffusivity [mm^2/s] from KD2 Pro instrument.

    Returns:
        A tuple of (time_grid, depth_grid, temp_grid), each a 2D NumPy array.
    """
    # Constants
    phase_frequency = 2 * np.pi / DAY_SECONDS # Phase Frequency | Units: 1 / s

    # Calculate Damping Depth
    thermal_diffusivity = thermal_diffusivity / 100 # Unit Conversion: cm^2 / s
    damp_depth = (2 * thermal_diffusivity / phase_frequency) ** 0.5 # Units: cm


    measurement_offset = DAY_SECONDS/len(air_temps)

    # Estimate sinusoidal parameters
    avg_temp = np.mean(air_temps)    # Units: °C
    thermal_amp = (max(air_temps) - min(air_temps)) / 2 # Units: °C

    # Find TimeLag
    min_idx = np.argmin(air_temps)
    timelag = min_idx*measurement_offset

    # Create time and depth grids
    times = [i*measurement_offset for i in len(air_temps)] # Units: s
    depths = np.linspace(max_depth/interpolations, max_depth, interpolations)
    time_grid, depth_grid = np.meshgrid(times, depths)

    # Generate predictions for the given times and depths
    temp_grid = avg_temp + \
        thermal_amp * np.exp(-depth_grid / damp_depth) * \
        np.sin(phase_frequency * (time_grid - timelag) - depth_grid / damp_depth)

    return time_grid, depth_grid, temp_grid


# EvapoTranspiration Models
def dalton(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        RH_min: np.ndarray,
        RH_max: np.ndarray,
        wind_speed: np.ndarray
    ) -> np.ndarray:
    """
    Computes Evapotranspiration using the Dalton model 
    
    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        RH_min:   The minimum daily relative humidity (range: 0.0-1.0)
        RH_max:   The maximum daily relative humidity (range: 0.0-1.0)
        wind_speed: The average daily wind speed in meters per second

    Returns:
        The Dalton models predictions for evapotranspiration clipped to a minimum of zero (mm/day)
    """
    # Ensures parameter saftey
    if not isinstance(temp_min, float) and not \
        (len(temp_min) == len(temp_max) == len(RH_min) == len(RH_max) == len(wind_speed)):
        raise ValueError("All inputs must be the same length")
    
    # Model calulations
    e_sat_min = __compute_vapor_sat_pressure(temp_min)
    e_sat_max = __compute_vapor_sat_pressure(temp_max)
    e_sat = (e_sat_min + e_sat_max)/2
    e_atm = (e_sat_min*(RH_max/100) + e_sat_max*(RH_min/100))/ 2
    PET = (3.648 + 0.7223*wind_speed)*(e_sat - e_atm)

    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return PET

def penman(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        RH_min: np.ndarray,
        RH_max: np.ndarray,
        wind_speed: np.ndarray
    ) -> np.ndarray:
    """
    Computes evapotranspiration using the Penman model

    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        RH_min:   The minimum daily relative humidity (range: 0.0-1.0)
        RH_max:   The maximum daily relative humidity (range: 0.0-1.0)
        wind_speed: The average daily wind speed in meters per second

    Returns:
        The Penman models predictions for evapotranspiration clipped to a minimum of zero
    """
    # Ensure parameter saftey
    if not isinstance(temp_min, float) and not \
        (len(temp_min) == len(temp_max) == len(RH_min) == len(RH_max) == len(wind_speed)):
        raise ValueError("All inputs must be the same length")
    
    # Model calulations
    e_sat_min = __compute_vapor_sat_pressure(temp_min)
    e_sat_max = __compute_vapor_sat_pressure(temp_max)
    e_sat = (e_sat_min + e_sat_max)/2
    e_atm = (e_sat_min*(RH_max/100) + e_sat_max*(RH_min/100))/ 2
    PET = (2.625 + 0.000479/wind_speed)*(e_sat - e_atm)

    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return PET

def romanenko(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        RH_min: np.ndarray,
        RH_max: np.ndarray,
    ) -> np.ndarray:
    """
    Potential evaporation model proposed by Romanenko in 1961
    
    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        RH_min:   The minimum daily relative humidity (range: 0.0-1.0)
        RH_max:   The maximum daily relative humidity (range: 0.0-1.0)

    Returns:
        The Romanenko models predictions for evapotranspiration clipped to a minimum of zero
    """
    # Ensure parameter saftey
    if not isinstance(temp_min, float) and not \
        (len(temp_min) == len(temp_max) == len(RH_min) == len(RH_max)):
        raise ValueError("All inputs must be the same length")

    # Model calulations
    temp_avg = (temp_min + temp_max)/2
    RH_avg = (RH_min + RH_max)/2
    PET = 0.00006*(25 + temp_avg)**2*(100 - RH_avg)

    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return PET

def jensen_haise(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        doy: np.ndarray,
        latitude: np.ndarray
    ) -> np.ndarray:
    """
    Computes evapotranspiration using the Jensen model
    
    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        doy: Day of year (0-365) where January 1st is 0 and 365
        latitude: Latitude in decimal degress. Where the northern hemisphere is 
            positive and the southern hemisphere is negative

    Returns:
        The Jensen-Haise models predictions for evapotranspiration clipped to a minimum of zero
    """
    # Ensure parameter saftey
    if not isinstance(temp_min, float) and not (len(temp_min) == len(temp_max)):
        raise ValueError("All inputs must be the same length")
    
    # Model Calculations
    Ra = __compute_solar_radiation(doy, latitude)
    T_avg = (temp_min + temp_max)/2
    PET = 0.0102 * (T_avg+3) * Ra

    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return PET

def hargreaves(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        doys: np.ndarray,
        latitude: float
    ) -> np.ndarray:
    """
    Computes evapotranspiration using the Hargreaves model
    
    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        doys: Array of Day of year's (0-365) where January 1st is 0 and 365
        latitude: Latitude in decimal degress. Where the northern hemisphere is 
            positive and the southern hemisphere is negative
        
    Returns:
        The Hargreaves models predictions for evapotranspiration clipped to a minimum of zero
    """
    # Ensure parameter saftey
    if not isinstance(temp_min, float) and not (len(temp_min) == len(temp_max)):
        raise ValueError("All inputs must be the same length")
    
    # Model Calulations
    Ra = __compute_solar_radiation(doys, latitude)
    T_avg = (temp_min + temp_max)/2
    PET = 0.0023 * Ra * (T_avg + 17.8) * (temp_max - temp_min)**0.5

    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return PET

def penman_monteith(
        temp_min: np.ndarray,
        temp_max: np.ndarray,
        RH_min: np.ndarray,
        RH_max: np.ndarray,
        p_min: np.ndarray,
        p_max: np.ndarray,
        wind_speed: np.ndarray,
        doy: np.ndarray,
        latitude: float,
        altitude: float,
        solar_rad: Optional[np.ndarray]=None,
        wind_height: int = 1.5,
    ) -> np.ndarray:
    """
    Computed evapotranspiration using the penman-monteith model
    
    Args:
        temp_min: The minimum daily temperature (°C)
        temp_max: The maximum daily temperature (°C)
        RH_min:   The minimum daily relative humidity (range: 0.0-1.0)
        RH_max:   The maximum daily relative humidity (range: 0.0-1.0)
        p_min:    The minimum daily atmospheric pressure in Pa
        p_max:    The maximum daily atmospheric pressure in Pa
        wind_speed: The average daily wind speed in meters per second
        doy:      Day of year (0-365) where January 1st is 0 and 365
        latitude: Latitude in decimal degress. Where the northern hemisphere is 
            positive and the southern hemisphere is negative
        altitude: Altitude of the location in meters
        wind_height: Height of measurment for wind speed
        
    Returns:
        The Hargreaves models predictions for evapotranspiration clipped to
            a minimum of zero
    """
    # Ensure parameter saftey
    if not isinstance(temp_min, float) and not (len(temp_min) == len(temp_max) == len(RH_min)\
                                     == len(RH_max) ==  len(wind_speed) == len(wind_speed)):
        raise ValueError("All inputs must be the same length")
    
    if solar_rad is not None:
        if len(solar_rad) != len(temp_min):
               raise ValueError("All inputs must be the same length")
    else:
        # Calculates solar radiation
        solar_rad = __compute_solar_radiation(doy, latitude)
    
    temp_avg = (temp_min + temp_max)/2
    atm_pressure = (p_min+p_max)/2 # Can be also obtained from weather station
    Cp = 0.001013; # Approx. 0.001013 for average atmospheric conditions
    gamma = 0.000665* (Cp * atm_pressure)

    # Wind speed Adjustment
    wind_speed_at_height = wind_speed * (4.87 / np.log((67.8 * wind_height) - 5.42))  # Eq. 47, FAO-56 wind height in [m]

    # Calculates air humidity and vapor pressure
    delta = 4098 * (0.6108 * np.exp(17.27 * temp_avg / (temp_avg  + 237.3)))
    delta = delta / (temp_avg  + 237.3)**2
    e_temp_max = 0.6108 * np.exp(17.27 * temp_max / (temp_max + 237.3)) # Eq. 11, //FAO-56
    e_temp_min = 0.6108 * np.exp(17.27 * temp_min / (temp_min + 237.3))
    e_saturation = (e_temp_max + e_temp_min) / 2
    e_actual = (e_temp_min * (RH_max / 100) + e_temp_max * (RH_min / 100)) / 2

    # Clear Sky Radiation: Rso (MJ/m2/day)
    Rso =  (0.75 + (2 * 10**-5) * altitude) * solar_rad  # Eq. 37, FAO-56

    # Rs/Rso = relative shortwave radiation (limited to <= 1.0)
    alpha = 0.23 # 0.23 for hypothetical grass reference crop
    Rns = (1 - alpha) * solar_rad # Eq. 38, FAO-56
    sigma  = 4.903 * 10**-9
    maxTempK = temp_max + 273.16
    minTempK = temp_min + 273.16


    # Eq. 39, FAO-56
    Rnl =  sigma * (maxTempK**4 + minTempK**4)
    Rnl = Rnl / 2 * (0.34 - 0.14 * np.sqrt(e_actual))
    Rnl = Rnl * (1.35 * (solar_rad / Rso) - 0.35) 
    Rn = Rns - Rnl # Eq. 40, FAO-56

    # Soil heat flux density
    soil_heat_flux = 0 # Eq. 42, FAO-56 G = 0 for daily time steps  [MJ/m2/day]

    # ETo calculation
    PET = 0.408 * delta * (Rn - soil_heat_flux)
    PET = PET + gamma * (900 / (temp_avg  + 273))  * wind_speed_at_height * (e_saturation - e_actual)
    PET = PET / (delta + gamma * (1 + 0.34 * wind_speed_at_height))


    # Ensures non-negative evapotranspiration
    PET = np.clip(PET, 0.0, None)

    return np.round(PET,2)

# Rain/Runoff Models
def model_runoff(rainfall: np.ndarray, curve_number: int = 75) -> pd.DataFrame:
    '''
    Uses Curve Number to estimate runoff from rainfall

    Args:
        precipitation: The daily amount of precipitation in millimeters
        curve_number: The curve number to use in the calculation

    Returns:
        The estimated runoff
    '''
    rainfall = rainfall / 25.4 # Unit Conversion | Units: In

    # Model Calulations
    runoff = np.zeros_like(rainfall)
    S02 = 1000 / curve_number - 10
    S005 = 1.33 * S02**1.15
    Lambda = 0.05
    Ia = S005 * Lambda
    idx = rainfall > Ia
    runoff[idx] = (rainfall[idx] - Ia)**2 / (rainfall[idx] - Ia + S005)

    return runoff * 25.4 # Unit Conversion | Units: mm

# Growing Degree Days
def model_gdd(
        temp_avg: np.ndarray,
        temp_base: float,
        temp_opt: Optional[float] = None,
        temp_upper: Optional[float] = None,
        time_duration: float = 1.0
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Model how many growing degree days have passed

    Models Growing Degree days using a minimum base temperature, an optional 
        optimal temperature, an optional maximum growing temperature, and 
        the average temperature over the recorded durations
        
    Args:
        temp_avg:  The average temperature over the duration_time (°C)
        temp_base: The minimum temperature a given crop will grow at (°C)
        temp_opt: The optimal temperature a given crop will grow (°C), 
                            above this temperature growing will slow linearly
        temp_upper: The maximum temperature a given crop will grow (°C)
        time_duration: The number of days that each temp_avg represents
    
    Returns:
        A tuple containing two np.ndarray's
        The first contains Growing Degree Days for each daily temperature
        The second contains the cummulative sum of Growing Degree Days for each daily temperature
    """
    # Validate parameter input
    if(time_duration <= 0):
        raise ValueError("The time duration must be positive")
    if temp_avg is None or (not np.isscalar(temp_avg) and len(temp_avg) == 0):
        raise ValueError("You must provide temperature averages, None provided")
    if temp_opt is None and temp_upper is not None or\
        temp_upper is None and temp_opt is not None:
        raise ValueError("You must provide both upper and optimal temperatures or neither")
    
    # Ensure the temp_avg is a numpy array
    temp_avg = np.asarray(temp_avg)
    
    # Simply GDD calculation
    if temp_opt is None and temp_upper is None:
        gdd_days = (temp_avg - temp_base)*time_duration
        gdd_days = np.maximum(0, gdd_days)
        return gdd_days, gdd_days.cumsum()

    # Initialize GDD array
    gdd = np.zeros_like(temp_avg)

    # Vectorized masks
    below_opt = temp_avg <= temp_opt
    above_opt = temp_avg > temp_opt 
    
    # Compute GDD where temp < temp_opt
    gdd[below_opt] = np.maximum( 0, (temp_avg[below_opt] - temp_base) * time_duration)
    
    # Compute GDD where temp >= temp_opt
    gdd_max = (temp_opt - temp_base) * time_duration
    gdd[above_opt] = np.maximum(0, 
        gdd_max * (temp_upper - temp_avg[above_opt]) / (temp_upper - temp_opt),)
    
    return gdd, gdd.cumsum()


# Photo Period Tools
def photoperiod_at_lat(lat: float, doy: np.ndarray) -> Tuple[np.array, np.array, np.array, np.array, np.array]:
    """
    Computes photoperiod for a given latitude and day of year. Not accurate near polar regions.

    Args:
        latitude: Latitude in decimal degress. Where the northern hemisphere is 
            positive and the southern hemisphere is negative
        doy: np.ndarray of the days of year (0-365) where January 1st is 0 and 365

    Returns:
        Photoperiod, daylight hours, at the given latitude for the given days.
        The angle of the sum below the horizon.
        The zenithal distance of the sun in degrees.
        The mean anomaly of the sun.
        Lambda.
        Delta.
    """
    # Convert latitude to radians
    lat_radians = np.radians(lat)
    
    # Angle of the sun below the horizon. Civil twilight is -4.76 degrees.
    light_intensity = 2.206 * 10**-3
    sun_angle = -4.76 - 1.03 * np.log(light_intensity) # Eq. [5].

    # Zenithal distance of the sun in degrees
    sun_zenithal_dist = np.radians(90 + sun_angle) # Eq. [6]. Value at sunrise and sunset.
    
    # Mean anomaly of the sun. It is a convenient uniform measure of 
    # how far around its orbit a body has progressed since pericenter.
    sun_mean_anomaly = 0.9856*doy - 3.251 # Eq. [4].
    
    # Declination of sun in degrees
    sun_declenation = sun_mean_anomaly + 1.916*np.sin(np.radians(sun_mean_anomaly)) + 0.020*np.sin(np.radians(2*sun_mean_anomaly)) + 282.565 # Eq. [3]. Lambda
    C = np.sin(np.radians(23.44)) # 23.44 degrees is the orbital plane of Earth around the Sun
    delta = np.arcsin(C*np.sin(np.radians(sun_declenation))) # Eq. [2].

    # Calculate daylength in hours, defining sec(x) = 1/cos(x)
    day_length = 2/15 * np.degrees( np.arccos( np.cos(sun_zenithal_dist) * (1/np.cos(lat_radians)) * (1/np.cos(delta)) - np.tan(lat_radians) * np.tan(delta) ) ) # Eq. [1].

    return day_length, sun_angle, sun_zenithal_dist, sun_mean_anomaly, sun_declenation, np.degrees(delta)
    
def photoperiod_on_day(lat: np.ndarray, doys: np.ndarray) -> Tuple[np.array, np.array, np.array, np.array, np.array]:
    """
    Computes photoperiod for a given near polar regions.

    Args:
        lat: Latitude in decimal degress. Where the northern hemisphere is positive and the southern hemisphere is negative.
        doys: The day of year (0-365) where January 1st is 0 and 365 to perform the calculation.

    Returns:
        Photoperiod, daylight hours, for the given latitudes on the given day.
        The angle of the sum below the horizon.
        The zenithal distance of the sun in degrees.
        The mean anomaly of the sun.
        The declination of the sun in degrees.
        Lambda from the equation.
        Delta from the equation.
    
    References:
        Keisling, T.C., 1982. Calculation of the Length of Day 1. Agronomy Journal, 74(4), pp.758-759.
    """
    # Convert latitude to radians and convert shapes
    lat = np.radians(np.asarray(lat)).reshape(-1, 1)  # shape (N, 1)
    doys = np.asarray(doys).reshape(1, -1) 
    
    # Angle of the sun below the horizon. Civil twilight is -4.76 degrees.
    light_intensity = 2.206 * 10**-3
    B = -4.76 - 1.03 * np.log(light_intensity) # Eq. [5].

    # Zenithal distance of the sun in degrees
    alpha = np.radians(90 + B) # Eq. [6]. Value at sunrise and sunset.
    
    # Mean anomaly of the sun. It is a convenient uniform measure of 
    # how far around its orbit a body has progressed since pericenter.
    M = 0.9856*doys - 3.251 # Eq. [4].
    
    # Declination of sun in degrees
    Lambda = M + 1.916*np.sin(np.radians(M)) + 0.020*np.sin(np.radians(2*M)) + 282.565 # Eq. [3]. Lambda
    C = np.sin(np.radians(23.44)) # 23.44 degrees is the orbital plane of Earth around the Sun
    delta = np.arcsin(C*np.sin(np.radians(Lambda))) # Eq. [2].

    # Calculate daylength in hours, defining sec(x) = 1/cos(x)
    P = 2/15 * np.degrees( np.arccos( np.cos(alpha) * (1/np.cos(lat)) * (1/np.cos(delta)) - np.tan(lat) * np.tan(delta) ) ) # Eq. [1].

    return P, B, alpha, M, Lambda, np.degrees(delta)


# Soil Water Flow Functions
def hydraulic_conductivity(
        sat_hydraulic_conductivity: float,
        air_entry_water_potential: float,
        volumetric_water_content: float,
        sat_water_content: float,
        b: float,
    ) -> float:
    """
    Estimates the hydraulic conductivity of soil.
    
    Args:
        sat_hydraulic_conductivity: The saturated conductivity of the soil  (kg * s / m^3).
        air_entry_water_potential: The air entry water potential of the soil  (J / kg).
        volumetric_water_content: Is the volumetric water content of the soil (m^3 / m^3).
        sat_water_content: Is the saturation water content of the soil (m^3 / m^3).
        b: Is the exponent of moisture release.

    Returns:
        The calculated hydraulic conductivity of the soil given the parameters ().
    """
    wetting_front_water_potential = air_entry_water_potential* (volumetric_water_content / sat_water_content) ** (-b)

    hydraulic_conductivity = sat_hydraulic_conductivity * (air_entry_water_potential / wetting_front_water_potential) ** (2 + 3/b)

    return hydraulic_conductivity

def cummulative_water_infiltration(
        water_vol_fraction: float,
        avg_sat_hydraulic_conductivity: float,
        infultration_water_potential: float,
        wetting_front_water_potential: float,
        max_time: int,
        interpolations: int=100
    ) -> np.array:
    """
    Calculates the cummulative vertical water infultration.
    
    Calculates the cummulative vertical water infultration for the given soil parameters.
        Calculations are made for equally spaced times from [0 to max_time].
    
    Args:
        water_vol_fraction: The Volume Fraction of water.
        avg_sat_hydraulic_conductivity: The average hydraulic conductivity of the wet soil (kg * s /m^3).
        infultration_water_potential: The infultration boudary's water potential (J / kg).
        wetting_front_water_potential: The wetting front's water potential (J / kg).
        max_time: The length of time to calcute the cummulative water infultration (s).
        interpolations: The number of interpolated times to calculate.
  
    Returns:
        A numpy array of length 'interpolations' representing the cummulative vertical water infultration over equal time segments.
        Where the first value represents the infultration after max_time/interpolations time,
            and the last vaule represents the infultration after max_time.
    """
    # Define Constants
    WATER_DENSITY = 1000 # Units: Kg / m^3
    LITTLE_G = 9.80665 # Units: m / s^2

    if water_vol_fraction < 0 or water_vol_fraction > 1.0:
        raise ValueError(f"The volume Fraction of water must be between [0,1], but {water_vol_fraction} was provided")

    # Calculate water infultration at each of the time
    times = np.linspace(max_time/interpolations, max_time, interpolations)
    water_inful = 2 * WATER_DENSITY * water_vol_fraction * avg_sat_hydraulic_conductivity * (infultration_water_potential - wetting_front_water_potential) * times
    water_inful =  water_inful ** 0.5 + LITTLE_G * avg_sat_hydraulic_conductivity * times

    return water_inful



